#!/usr/bin/env python
"""
Wrapper for hacl-star XK_Noise

   SPDX-FileCopyrightText: 2024, Marsiske Stefan
   SPDX-License-Identifier: LGPL-3.0-or-later

Copyright (c) 2024, Marsiske Stefan.
All rights reserved.

  This file is part of liboprf.

  liboprf is free software: you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation, either version 3 of
  the License, or (at your option) any later version.

  liboprf is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with liboprf. If not, see <http://www.gnu.org/licenses/>.

"""


import ctypes
import ctypes.util
from ctypes import c_void_p, c_ubyte, c_uint32, c_char, c_size_t, POINTER, byref

lib = ctypes.cdll.LoadLibrary(ctypes.util.find_library('oprf-noiseXK')
                              or ctypes.util.find_library('liboprf-noiseXK'))
if not lib._name:
    raise ValueError('Unable to find liboprf-noiseXK')
libc = ctypes.cdll.LoadLibrary(ctypes.util.find_library('c') or ctypes.util.find_library('libc'))
if not libc._name:
    raise ValueError('Unable to find libc')

KEYSIZE = 32
NOISE_XK_CONF_ZERO = 0
NOISE_XK_AUTH_KNOWN_SENDER_NO_KCI = 2
NOISE_XK_CONF_STRONG_FORWARD_SECRECY = 5

def __check(code):
    if code != 0:
        raise ValueError

lib.Noise_XK_device_add_peer.restype = c_void_p
lib.Noise_XK_device_add_peer.argtypes = [c_void_p, c_void_p, ctypes.c_char_p]
def add_peer(device, name, key):
    return lib.Noise_XK_device_add_peer(device, name, key)

def pubkey(privkey):
    pubkey = ctypes.create_string_buffer(KEYSIZE)
    lib.Noise_XK_dh_secret_to_public(pubkey, privkey)
    return pubkey.raw

lib.Noise_XK_device_create.restype = c_void_p
def create_device(prologue, name, privkey):
    srlz_key = b'\x00'*KEYSIZE
    return lib.Noise_XK_device_create(len(prologue), prologue, name, srlz_key, privkey)

lib.Noise_XK_peer_get_id.restype = c_void_p
lib.Noise_XK_peer_get_id.argtypes = [c_void_p]
def get_peerid(peer):
    return lib.Noise_XK_peer_get_id(peer)

lib.Noise_XK_session_create_initiator.restype = c_void_p
lib.Noise_XK_session_create_initiator.argtypes = [c_void_p, c_void_p]
def create_session_initiator(device, peerid):
    return lib.Noise_XK_session_create_initiator(device, peerid)


lib.Noise_XK_session_create_initiator.restype = c_void_p
lib.Noise_XK_session_create_initiator.argtypes = [c_void_p, c_void_p]
def create_session_initiator(device, peerid):
    res = lib.Noise_XK_session_create_initiator(device, peerid)
    if res == 0: raise ValueError
    return res

lib.Noise_XK_session_create_responder.restype = c_void_p
lib.Noise_XK_session_create_responder.argtypes = [c_void_p]
def create_session_responder(device):
    res = lib.Noise_XK_session_create_responder(device)
    if res == 0: raise ValueError
    return res

lib.Noise_XK_pack_message_with_conf_level.restype = c_void_p
lib.Noise_XK_session_write.argtypes = [c_void_p, c_void_p, POINTER(c_uint32), POINTER(POINTER(c_ubyte))]
lib.Noise_XK_encap_message_p_free.argtypes = [c_void_p]
def initiator_1st_msg(session):
    encap_msg = lib.Noise_XK_pack_message_with_conf_level(0, 0, 0);
    msg_len = c_uint32()
    msg = POINTER(c_ubyte)()
    if 0!=lib.Noise_XK_session_write(encap_msg, session, byref(msg_len), byref(msg)):
        raise ValueError
    lib.Noise_XK_encap_message_p_free(encap_msg)
    res = bytes(msg[i] for i in range(msg_len.value))
    if msg_len.value > 0:
        libc.free(msg)
    return res

# Noise_XK_session_read(&encap_msg, bob_session, cipher_msg_len, cipher_msg);
lib.Noise_XK_session_read.argtypes = [POINTER(c_void_p), c_void_p, c_uint32, POINTER(c_ubyte)]
# Noise_XK_unpack_message_with_auth_level(&plain_msg_len, &plain_msg, NOISE_XK_AUTH_ZERO, encap_msg),
def responder_1st_msg(session, msg):
    encap_msg = c_void_p()
    msg = (c_ubyte * len(msg)).from_buffer(bytearray(msg))
    msg_len = c_uint32(len(msg))
    if 0 != lib.Noise_XK_session_read(byref(encap_msg), session, msg_len, msg):
        raise ValueError
    plain_msg_len = c_uint32()
    plain_msg = POINTER(c_ubyte)()
    if not lib.Noise_XK_unpack_message_with_auth_level(byref(plain_msg_len), byref(plain_msg), 0, encap_msg):
        raise ValueError
    lib.Noise_XK_encap_message_p_free(encap_msg)
    if plain_msg_len.value > 0:
        libc.free(plain_msg)
    return initiator_1st_msg(session)

def initiator_handshake_finish(session, msg):
    encap_msg = c_void_p()
    msg = (c_ubyte * len(msg)).from_buffer(bytearray(msg))
    msg_len = c_uint32(len(msg))
    if 0 != lib.Noise_XK_session_read(byref(encap_msg), session, msg_len, msg):
        raise ValueError
    plain_msg_len = c_uint32()
    plain_msg = POINTER(c_ubyte)()
    if not lib.Noise_XK_unpack_message_with_auth_level(byref(plain_msg_len), byref(plain_msg), 0, encap_msg):
        raise ValueError
    lib.Noise_XK_encap_message_p_free(encap_msg)
    if plain_msg_len.value > 0:
        libc.free(plain_msg)

def send_msg(session, msg):
    if isinstance(msg, str): msg = msg.encode('utf8')
    encap_msg = lib.Noise_XK_pack_message_with_conf_level(NOISE_XK_CONF_STRONG_FORWARD_SECRECY, len(msg), msg);
    ct_len = c_uint32()
    ct = POINTER(c_ubyte)()
    if 0!=lib.Noise_XK_session_write(encap_msg, session, byref(ct_len), byref(ct)):
        raise ValueError
    lib.Noise_XK_encap_message_p_free(encap_msg)
    res = bytes(ct[:ct_len.value])
    if ct_len.value > 0:
        libc.free(ct)
    return res

def read_msg(session, msg):
    encap_msg = c_void_p()
    u_bytes = (c_ubyte * (len(msg)))()
    u_bytes[:] = msg
    if 0 != lib.Noise_XK_session_read(byref(encap_msg), session, len(msg), u_bytes):
        raise ValueError
    plain_msg_len = c_uint32()
    plain_msg = POINTER(c_ubyte)()
    if not lib.Noise_XK_unpack_message_with_auth_level(byref(plain_msg_len), byref(plain_msg),
                                                              NOISE_XK_AUTH_KNOWN_SENDER_NO_KCI, encap_msg):
        raise ValueError
    lib.Noise_XK_encap_message_p_free(encap_msg)
    res = bytes(plain_msg[i] for i in range(plain_msg_len.value))
    if plain_msg_len.value > 0:
        libc.free(plain_msg)
    return res

lib.Noise_XK_session_get_peer_id.restype = c_uint32
lib.Noise_XK_session_get_peer_id.argtypes = [c_void_p]
lib.Noise_XK_device_lookup_peer_by_id.restype = c_void_p
lib.Noise_XK_device_lookup_peer_by_id.argtypes = [c_void_p, c_uint32]
lib.Noise_XK_peer_get_static.argtypes = [(c_char * 32), c_void_p]
def get_pubkey(session, device):
    peerid = lib.Noise_XK_session_get_peer_id(session)
    peer = lib.Noise_XK_device_lookup_peer_by_id(device, peerid);
    pubkey = ctypes.create_string_buffer(KEYSIZE)
    lib.Noise_XK_peer_get_static(pubkey, peer);
    return pubkey.raw

def initiator_session(initiator_privkey, responder_pubkey, iname=None,
                      rname=None, dst=None):
    if dst is None:
        dst = b"liboprf-noiseXK"
    if iname is None:
        iname = b"initiator"
    if rname is None:
        rname = b"responder"
    #initiator_pubkey = pubkey(initiator_privkey)
    dev = create_device(dst, iname, initiator_privkey)
    peer = add_peer(dev, rname, responder_pubkey)
    peerid = get_peerid(peer)

    session = create_session_initiator(dev, peerid)
    msg = initiator_1st_msg(session)
    return session, msg

libc.malloc.restype = POINTER(c_ubyte)
def responder_session(responder_privkey, auth_keys, msg, dst=None, name=None):
    if dst is None:
        dst = b"liboprf-noiseXK"
    if name is None:
        name = b"responder"
    #responder_pubkey = pubkey(responder_privkey)
    dev = create_device(dst, name, responder_privkey)
    for key, peer in auth_keys:
        add_peer(dev,peer,key)
    session = create_session_responder(dev)
    msg = responder_1st_msg(session, msg)
    return session, msg

def initiator_session_complete(session, msg):
    return initiator_handshake_finish(session, msg)

def test():
    from binascii import unhexlify, hexlify
    # low level
    alice_privkey = unhexlify("c3da55379de9c6908e94ea4df28d084f32eccf03491c71f754b4075577a28552")
    alice_pubkey = pubkey(alice_privkey)
    bob_privkey = unhexlify("c3da55379de9c6908e94ea4df28d084f32eccf03491c71f754b4075577a28552")
    bob_pubkey = pubkey(bob_privkey)

    adev = create_device("liboprf-noiseXK test", "Alice", alice_privkey)
    bpeer = add_peer(adev, "Bob", bob_pubkey)
    bobid = get_peerid(bpeer)

    bdev = create_device("liboprf-noiseXK test", "Bob", bob_privkey)
    add_peer(bdev, "Alice", alice_pubkey)

    asession = create_session_initiator(adev, bobid)
    bsession = create_session_responder(bdev)

    msg = initiator_1st_msg(asession)

    msg = responder_1st_msg(bsession, msg)

    initiator_handshake_finish(asession, msg)

    ct = send_msg(asession, "hello bob!")
    pt = read_msg(bsession, ct)

    peer_pk = get_pubkey(bsession, bdev)
    print(hexlify(peer_pk))

    print(pt)
    ct = send_msg(bsession, "hello alice!")
    pt = read_msg(asession, ct)
    print(pt)

    # high-level
    a2session, msg = initiator_session(alice_privkey, bob_pubkey)
    b2session, msg = responder_session(bob_privkey, [(alice_pubkey, "Alice")], msg)
    initiator_session_complete(a2session, msg)

    ct = send_msg(a2session, "hello bob!")
    pt = read_msg(b2session, ct)
    print(pt)
    ct = send_msg(b2session, "hello alice!")
    pt = read_msg(a2session, ct)
    print(pt)

    for _ in range(1000):
        if ct[0] % 2 == 0:
            sender = a2session
            receiver = b2session
        else:
            sender = b2session
            receiver = a2session
        message = ct[:16+(ct[1]>>4)] * (ct[1] & 0xf)
        ct = send_msg(sender, message)
        pt = read_msg(receiver, ct)
        assert(pt == message)

if __name__ == '__main__':
    test()



#!/usr/bin/env python

from pyoprf import keygen, create_shares, blind, evaluate, unblind, thresholdmult
from pysodium import randombytes, crypto_core_ristretto255_from_hash, crypto_generichash, crypto_core_ristretto255_add

k = keygen()
shares = create_shares(k, 5, 3)

zero_shares = create_shares(bytes([0]*32), 5, 3)

r, alpha = blind(b"test")

ssid_S = randombytes(32)
betas = []
for ki, zi in zip(shares,zero_shares):
    h2 = evaluate(
        zi[1:],
        crypto_core_ristretto255_from_hash(crypto_generichash(ssid_S + alpha, outlen=64)),
        )
    beta = evaluate(ki[1:], alpha)
    betas.append(ki[:1]+crypto_core_ristretto255_add(beta, h2))

# normal 2hashdh(k,"test")
beta = evaluate(k, alpha)
Nt0 = unblind(r, beta)
print(Nt0)
beta = thresholdmult(betas[:3])
Nt1 = unblind(r, beta)
print(Nt1)
assert Nt0 == Nt1



#!/usr/bin/env python3

import json, sys
from itertools import zip_longest # for Python 3.x

def split_by_n(iterable, n):
    return zip_longest(*[iter(iterable)]*n, fillvalue='')

# src: ht
vectors = """
{
	"groupDST": "48617368546f47726f75702d4f50524656312d002d72697374726574746f3235352d534841353132",
	"hash": "SHA512",
	"identifier": "ristretto255-SHA512",
	"keyInfo": "74657374206b6579",
	"mode": 0,
	"seed": "a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3a3",
	"skSm": "5ebcea5ee37023ccb9fc2d2019f9d7737be85591ae8652ffa9ef0f4d37063b0e",
	"vectors": [
	  {
		"Batch": 1,
		"Blind": "64d37aed22a27f5191de1c1d69fadb899d8862b58eb4220029e036ec4c1f6706",
		"BlindedElement": "609a0ae68c15a3cf6903766461307e5c8bb2f95e7e6550e1ffa2dc99e412803c",
		"EvaluationElement": "7ec6578ae5120958eb2db1745758ff379e77cb64fe77b0b2d8cc917ea0869c7e",
		"Input": "00",
		"Output": "527759c3d9366f277d8c6020418d96bb393ba2afb20ff90df23fb7708264e2f3ab9135e3bd69955851de4b1f9fe8a0973396719b7912ba9ee8aa7d0b5e24bcf6"
	  },
	  {
		"Batch": 1,
		"Blind": "64d37aed22a27f5191de1c1d69fadb899d8862b58eb4220029e036ec4c1f6706",
		"BlindedElement": "da27ef466870f5f15296299850aa088629945a17d1f5b7f5ff043f76b3c06418",
		"EvaluationElement": "b4cbf5a4f1eeda5a63ce7b77c7d23f461db3fcab0dd28e4e17cecb5c90d02c25",
		"Input": "5a5a5a5a5a5a5a5a5a5a5a5a5a5a5a5a5a",
		"Output": "f4a74c9c592497375e796aa837e907b1a045d34306a749db9f34221f7e750cb4f2a6413a6bf6fa5e19ba6348eb673934a722a7ede2e7621306d18951e7cf2c73"
	  }
	]
}
"""

def toC(k, v):
  print(f"#define {k.lower()}_len {len(v)//2}")
  print(
      f"const uint8_t {k.lower()}[{k.lower()}_len] = {{\n   %s}};\n" % ",\n   ".join(
          (", ".join((c for c in line if c)) for line in split_by_n(
              (f"0x{x[0]}{x[1]}" for x in split_by_n(v,2))
              ,8))
      ))

vex = json.loads(vectors)
print("// this file has been automatically generated using testvecs2h.py")

if sys.argv[1] == 'cfrg_oprf_test_vectors.h':
   # run this if there is a change in the values of the test vectors
   # ./testvecs2h.py >cfrg_oprf_test_vectors.h

   print("#ifndef cfrg_test_vectors_h\n#define cfrg_test_vectors_h\n")
   print("#include <stdint.h>\n")

   toC("sks", vex['skSm'])

   for tc in range(2):
       for k, v in vex['vectors'][tc].items():
           if k == "Batch": continue
           print(f"#define tc{tc}_{k.lower()}_len {len(v)//2}")
           print(
               f"const uint8_t tc{tc}_{k.lower()}[tc{tc}_{k.lower()}_len] = {{\n   %s}};\n" % ",\n   ".join(
                   (", ".join((c for c in line if c)) for line in split_by_n(
                       (f"0x{x[0]}{x[1]}" for x in split_by_n(v,2))
                       ,8))
               ))

elif sys.argv[1] == 'cfrg_oprf_test_vector_decl.h':
   # only run this code below if there is a change in the keys of the test vectors
   # ./testvecs2h.py >cfrg_oprf_test_vector_decl.h

   print("#ifndef cfrg_test_vector_decl_h\n#define cfrg_test_vector_decl_h\n")
   print("#include <stdint.h>\n")
   for tc in range(2):
      for k, v in vex['vectors'][tc].items():
         if k == "Batch": continue
         print(f"#define tc{tc}_{k.lower()}_len {len(v)//2}")
         print(f"extern const uint8_t tc{tc}_{k.lower()}[tc{tc}_{k.lower()}_len];\n")
else:
    sys.exit(-1)

print("""
#if(TC==0)
    #define input                 tc0_input
    #define input_len             tc0_input_len
    #define blind_registration    tc0_blind
    #define blind_login           tc0_blind
    #define blind_len             tc0_blind_len
    #define blinded_element       tc0_blindedelement
    #define blindedelement_len    tc0_blindedelement_len
    #define evaluationelement     tc0_evaluationelement
    #define evaluationelement_len tc0_evaluationelement_len
    #define output                tc0_output
    #define output_len            tc0_output_len
#else
    #define input                 tc1_input
    #define input_len             tc1_input_len
    #define blind_registration    tc1_blind
    #define blind_login           tc1_blind
    #define blind_len             tc1_blind_len
    #define blinded_element       tc1_blindedelement
    #define blindedelement_len    tc1_blindedelement_len
    #define evaluationelement     tc1_evaluationelement
    #define evaluationelement_len tc1_evaluationelement_len
    #define output                tc1_output
    #define output_len            tc1_output_len
#endif""")
print("#endif")



#!/usr/bin/env python
"""
Wrapper for liboprf library

   SPDX-FileCopyrightText: 2023, Marsiske Stefan
   SPDX-License-Identifier: LGPL-3.0-or-later

Copyright (c) 2023, Marsiske Stefan.
All rights reserved.

  This file is part of liboprf.

  liboprf is free software: you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation, either version 3 of
  the License, or (at your option) any later version.

  liboprf is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with liboprf. If not, see <http://www.gnu.org/licenses/>.

"""

import ctypes
import ctypes.util
import pysodium, os
import platform
from typing import List, Tuple
from itertools import zip_longest

if "BYZANTINE_DKG" in os.environ:
    liboprf = ctypes.cdll.LoadLibrary(os.environ['BYZANTINE_DKG'])
else:
    liboprf = ctypes.cdll.LoadLibrary(ctypes.util.find_library('oprf') or ctypes.util.find_library('liboprf'))

if not liboprf._name:
    raise ValueError('Unable to find liboprf')

def split_by_n(iterable, n):
    return list(zip_longest(*[iter(iterable)]*n, fillvalue=''))

def __check(code):
    if code != 0:
        raise ValueError(f"error: {code}")

# (CFRG/IRTF) OPRF section

OPRF_BYTES=64

# This function generates an OPRF private key.
#
# This is almost the KeyGen OPRF function defined in the RFC: since
# this lib does not implement V oprf, we don't need a pubkey and so
# we don't bother with all that is related.
#
# @param [out] k - the per-user OPRF private key
# void oprf_KeyGen(uint8_t kU[crypto_core_ristretto255_SCALARBYTES]);
def keygen() -> bytes:
    k = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_SCALARBYTES)
    liboprf.oprf_KeyGen(k)
    return k.raw


# This function converts input x into an element of the OPRF group, randomizes it
# by some scalar r, producing blinded, and outputs (r, blinded).
#
# This is the Blind OPRF function defined in the RFC.
#
# @param [in] x - the input value to blind
# @param [out] r - an OPRF scalar value used for randomization
# @param [out] blinded - a serialized OPRF group element, a byte array of fixed length,
# the blinded version of x, an input to oprf_Evaluate
# @return The function raises a ValueError if there is something wrong with the inputs.
#
#int oprf_Blind(const uint8_t *x, const uint16_t x_len,
#               uint8_t r[crypto_core_ristretto255_SCALARBYTES],
#               uint8_t blinded[crypto_core_ristretto255_BYTES]);
def blind(x: bytes) -> (bytes, bytes):
    r =       ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_SCALARBYTES)
    blinded = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_BYTES)
    __check(liboprf.oprf_Blind(x, ctypes.c_size_t(len(x)), r, blinded))
    return r.raw, blinded.raw


# This function evaluates input element blinded using private key k, yielding output
# element Z.
#
# This is the Evaluate OPRF function defined in the RFC.
#
# @param [in] key - a private key - the output of keygen()
# @param [in] blinded - a serialized OPRF group element, a byte array
#                     of fixed length, an output of blind()
# @param [out] Z - a serialized OPRF group element, a byte array of fixed
#                     length, an input to oprf_Unblind
# @return The function raises a ValueError if there is something wrong with the inputs.
#int oprf_Evaluate(const uint8_t k[crypto_core_ristretto255_SCALARBYTES],
#                  const uint8_t blinded[crypto_core_ristretto255_BYTES],
#                  uint8_t Z[crypto_core_ristretto255_BYTES]);
def evaluate(key: bytes, blinded: bytes) -> bytes:
    if len(key) != pysodium.crypto_core_ristretto255_SCALARBYTES:
        raise ValueError("key has incorrect length")
    if not isinstance(key, bytes):
        raise ValueError("key is not of type bytes")
    if len(blinded) != pysodium.crypto_core_ristretto255_BYTES:
        raise ValueError("blinded param has incorrect length")
    if not isinstance(blinded, bytes):
        raise ValueError("blinded is not of type bytes")
    Z = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_BYTES)
    __check(liboprf.oprf_Evaluate(key, blinded, Z))
    return Z.raw


# This function removes random scalar r from Z, yielding output N.
#
# This is the Unblind OPRF function defined in the RFC.
#
# If you do not call finalize() on the result the output is equivalent
# to the OPRF protcol we refer to as HashDH - this protocol retains
# the algebraic structure of the value, and has weaker security
# guarantees, than the full 2HashDH which is equivalent to running
# finalize on the output of blind(). The hashDH variant is not
# explicitly specified by the CFRG/IRTF specification. This hashDH
# variant has one property that makes it interesting: it is an
# updateable OPRF - that is if the server updates their key, they can
# calculate a public delta value, that can be applied by the client to
# the output of blind() and the result will be as if the client and
# the server run the OPRF protocol with the original input and the new
# key. It is important to note that the delta value is not sensitive,
# and can be public.
#
# @param [in] r - an OPRF scalar value used for randomization in oprf_Blind
# @param [in] Z - a serialized OPRF group element, a byte array of fixed length,
#                 an output of oprf_Evaluate
# @param [out] N - a serialized OPRF group element with random scalar r removed,
#                 a byte array of fixed length, an input to oprf_Finalize
# @return The function raises a ValueError if there is something wrong with the inputs.
#int oprf_Unblind(const uint8_t r[crypto_core_ristretto255_SCALARBYTES],
#                 const uint8_t Z[crypto_core_ristretto255_BYTES],
#                 uint8_t N[crypto_core_ristretto255_BYTES]);
def unblind(r: bytes, Z: bytes) -> bytes:
    if len(r) != pysodium.crypto_core_ristretto255_SCALARBYTES:
        raise ValueError("param r has incorrect length")
    if not isinstance(r, bytes):
        raise ValueError("param r is not of type bytes")
    if len(Z) != pysodium.crypto_core_ristretto255_BYTES:
        raise ValueError("param Z has incorrect length")
    if not isinstance(Z, bytes):
        raise ValueError("param Z is not of type bytes")
    N = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_BYTES)
    __check(liboprf.oprf_Unblind(r, Z, N))
    return N.raw

# This function computes the OPRF output using input x, N, and domain
# separation tag info.
#
# This is the Finalize OPRF function defined in the RFC.
#
# @param [in] x - a value used to compute OPRF (the same value that
#                 was used as input to be blinded)
# @param [in] N - a serialized OPRF group element, a byte array of fixed length,
#                 an output of oprf_Unblind
# @param [out] y - an OPRF output
# @return The function raises a ValueError if there is something wrong with the inputs.
#int oprf_Finalize(const uint8_t *x, const uint16_t x_len,
#                  const uint8_t N[crypto_core_ristretto255_BYTES],
#                  uint8_t rwdU[OPRF_BYTES]);
def finalize(x: bytes, N: bytes) -> bytes:
    if len(N) != pysodium.crypto_core_ristretto255_BYTES:
        raise ValueError("param N has incorrect length")
    if not isinstance(N, bytes):
        raise ValueError("param N is not of type bytes")
    y = ctypes.create_string_buffer(OPRF_BYTES)
    __check(liboprf.oprf_Finalize(x, ctypes.c_size_t(len(x)), N, y))
    return y.raw

# This function combines unblind() and finalize() as a convenience
def unblind_finalize(r: bytes, Z: bytes, x: bytes) -> bytes:
    return finalize(x, unblind(r,Z))

# TOPRF section

TOPRF_Share_BYTES=pysodium.crypto_core_ristretto255_SCALARBYTES+1
TOPRF_Part_BYTES=pysodium.crypto_core_ristretto255_BYTES+1

# This function calculates a lagrange coefficient based on the index
# and the indexes of the other contributing shareholders.
#
# @param [in] index - the index of the shareholder whose lagrange
#             coefficient we're calculating, must be greater than 0
#
# @param [in] peers - list of the shares that contribute to the reconstruction
#
# @param [out] result - the lagrange coefficient
#void coeff(const int index, const int peers_len, const uint8_t peers[peers_len], uint8_t result[crypto_scalarmult_ristretto255_SCALARBYTES]);
def coeff(index: int, peers: list) -> bytes:
    if index < 1: raise ValueError("index must be positive integer")
    if len(peers) < 2: raise ValueError("peers must be a list of at least 2 integers")
    peers_len=ctypes.c_size_t(len(peers))
    c = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_SCALARBYTES)
    liboprf.coeff(index, peers_len, peers, c)
    return c.raw


# This function creates shares of secret in a (threshold, n) scheme
# over the curve ristretto255
#
# @param [in] secret - the scalar value to be secretly shared
#
# @param [in] n - the number of shares created
#
# @param [in] threshold - the threshold needed to reconstruct the secret
#
# @param [out] shares - n shares
#
# @return The function raises a ValueError if there is something wrong with the inputs.
#void toprf_create_shares(const uint8_t secret[crypto_core_ristretto255_SCALARBYTES],
#                   const uint8_t n,
#                   const uint8_t threshold,
#                   uint8_t shares[n][TOPRF_Share_BYTES]);
bytes_list_t = List[bytes]
def create_shares(secret: bytes, n: int, t: int) -> bytes_list_t:
    if len(secret) != pysodium.crypto_core_ristretto255_SCALARBYTES:
        raise ValueError("secret has incorrect length")
    if not isinstance(secret, bytes):
        raise ValueError("secret is not of type bytes")
    if n < t:
        raise ValueError("t cannot be bigger than n")
    if t < 2:
        raise ValueError("t must be bigger than 1")
    shares = ctypes.create_string_buffer(n*TOPRF_Share_BYTES)
    __check(liboprf.toprf_create_shares(secret, n, t, shares))
    return tuple([bytes(s) for s in split_by_n(shares.raw, TOPRF_Share_BYTES)])


# This function recovers the secret in the exponent using lagrange interpolation
# over the curve ristretto255
#
# The shareholders are not aware if they are contributing to a
# threshold or non-threshold oprf evaluation, from their perspective
# nothing changes in this approach.
#
# @param [in] responses - is an array of shares (k_i) multiplied by a
#        point (P) on the r255 curve
#
# @param [in] responses_len - the number of elements in the response array
#
# @param [out] result - the reconstructed value of P multipled by k
#
# @return The function raises a ValueError if there is something wrong with the inputs.
#int toprf_thresholdmult(const size_t response_len,
#                        const uint8_t responses[response_len][TOPRF_Part_BYTES],
#                        uint8_t result[crypto_scalarmult_ristretto255_BYTES]);
def thresholdmult(responses: bytes_list_t) -> bytes:
    if len(responses) < 2: raise ValueError("responses must be a list of at least 2 integers")
    if not all(isinstance(r,bytes) for r in responses):
        raise ValueError("at least one of the responses is not of type bytes")
    if not all(len(r)==TOPRF_Part_BYTES for r in responses):
        raise ValueError("at least one of the responses is not of correct size")
    responses_len=ctypes.c_size_t(len(responses))
    responses_buf = ctypes.create_string_buffer(b''.join(responses))
    result = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_BYTES)
    __check(liboprf.toprf_thresholdmult(responses_len, responses_buf, result))
    return result.raw


# This function is the efficient threshold version of oprf_Evaluate.
#
# This function needs to know in advance the indexes of all the
# shares that will be combined later in the toprf_thresholdcombine() function.
# by doing so this reduces the total costs and distributes them to the shareholders.
#
# @param [in] k - a private key (for OPAQUE, this is kU, the user's
#        OPRF private key)
#
# @param [in] blinded - a serialized OPRF group element, a byte array
#         of fixed length, an output of oprf_Blind (for OPAQUE, this
#         is the blinded pwdU, the user's password)
#
# @param [in] self - the index of the current shareholder
#
# @param [in] indexes - the indexes of the all the shareholders
#        contributing to this oprf evaluation,
#
# @param [in] index_len - the length of the indexes array,
#
# @param [out] Z - a serialized OPRF group element, a byte array of fixed length,
#        an input to oprf_Unblind
#
# @return The function raises a ValueError if there is something wrong with the inputs.
#int toprf_Evaluate(const uint8_t k[TOPRF_Share_BYTES],
#                   const uint8_t blinded[crypto_core_ristretto255_BYTES],
#                   const uint8_t self, const uint8_t *indexes, const uint16_t index_len,
#                   uint8_t Z[TOPRF_Part_BYTES]);
def threshold_evaluate(k: bytes, blinded: bytes, self: int, indexes: list) -> bytes:
    if len(k) != TOPRF_Share_BYTES:
        raise ValueError("param k has incorrect length")
    if not isinstance(k, bytes):
        raise ValueError("param k is not of type bytes")
    if len(blinded) != pysodium.crypto_core_ristretto255_BYTES:
        raise ValueError("blinded param has incorrect length")
    if not isinstance(blinded, bytes):
        raise ValueError("blinded is not of type bytes")
    if(self>255 or self<1):
        raise ValueError("self outside valid range")
    if(not all(i>0 and i<256 for i in indexes)):
        raise ValueError("index(es) outside valid range")
    index_len=ctypes.c_uint16(len(indexes))

    indexes_buf=ctypes.create_string_buffer(bytes(indexes))
    Z = ctypes.create_string_buffer(TOPRF_Part_BYTES)

    __check(liboprf.toprf_Evaluate(k, blinded, self, indexes_buf, index_len, Z))
    return Z.raw

# This function is combines the results of the toprf_Evaluate()
# function to recover the shared secret in the exponent.
#
# @param [in] responses - is an array of shares (k_i) multiplied by a point (P) on the r255 curve
#
# @param [in] responses_len - the number of elements in the response array
#
# @param [out] result - the reconstructed value of P multipled by k
#
# @return The function raises a ValueError if there is something wrong with the inputs.
#void toprf_thresholdcombine(const size_t response_len,
#                            const uint8_t _responses[response_len][TOPRF_Part_BYTES],
#                            uint8_t result[crypto_scalarmult_ristretto255_BYTES]);
def threshold_combine(responses: bytes_list_t) -> bytes:
    if len(responses) < 2: raise ValueError("responses must be a list of at least 2 integers")
    if not all(isinstance(r,bytes) for r in responses):
        raise ValueError("at least one of the responses is not of type bytes")
    if not all(len(r)==TOPRF_Part_BYTES for r in responses):
        raise ValueError("at least one of the responses is not of correct size")
    responses_len=ctypes.c_size_t(len(responses))
    responses_buf = ctypes.create_string_buffer(b''.join(responses))
    result = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_BYTES)

    __check(liboprf.toprf_thresholdcombine(responses_len, responses_buf, result))
    return result.raw

#int toprf_3hashtdh(const uint8_t k[TOPRF_Share_BYTES],
#                   const uint8_t z[TOPRF_Share_BYTES],
#                   const uint8_t alpha[crypto_core_ristretto255_BYTES],
#                   const uint8_t *ssid_S, const uint16_t ssid_S_len,
#                   uint8_t beta[TOPRF_Part_BYTES]);
def _3hashtdh(k: bytes, z: bytes, alpha: bytes, ssid_S: bytes) -> bytes:
    if len(k) != TOPRF_Share_BYTES:
        raise ValueError("param k has incorrect length")
    if not isinstance(k, bytes):
        raise ValueError("param k is not of type bytes")
    if len(z) != TOPRF_Share_BYTES:
        raise ValueError("param z has incorrect length")
    if not isinstance(z, bytes):
        raise ValueError("param z is not of type bytes")
    if len(alpha) != pysodium.crypto_core_ristretto255_BYTES:
        raise ValueError("alpha param has incorrect length")
    if not isinstance(alpha, bytes):
        raise ValueError("alpha is not of type bytes")
    if not isinstance(ssid_S, bytes):
        raise ValueError("ssid_S is not of type bytes")
    if len(ssid_S) > (1<<16)-1:
        raise ValueError("ssid_S is too long")

    ssid_S_len=ctypes.c_uint16(len(ssid_S))
    beta = ctypes.create_string_buffer(TOPRF_Part_BYTES)
    __check(liboprf.toprf_3hashtdh(k, z, alpha, ssid_S, ssid_S_len, beta))
    return beta.raw

# todo documentation!
#int dkg_start(const uint8_t n,
#              const uint8_t threshold,
#              uint8_t commitment_hash[dkg_hash_BYTES],
#              uint8_t commitments[dkg_commitment_BYTES(threshold)],
#              TOPRF_Share shares[n]);
def dkg_start(n : int, t : int) -> (bytes, bytes, bytes_list_t):
    if n < t:
        raise ValueError("t cannot be bigger than n")
    if t < 2:
        raise ValueError("t must be bigger than 1")
    shares = ctypes.create_string_buffer(n*TOPRF_Share_BYTES)
    commitments = ctypes.create_string_buffer(t*pysodium.crypto_core_ristretto255_BYTES)

    __check(liboprf.dkg_start(n, t, commitments, shares))

    shares = tuple([bytes(s) for s in split_by_n(shares.raw, TOPRF_Share_BYTES)])
    return commitments.raw, shares

#int dkg_verify_commitments(const uint8_t n,
#                           const uint8_t threshold,
#                           const uint8_t self,
#                           const uint8_t commitments[n][threshold*crypto_core_ristretto255_BYTES],
#                           const TOPRF_Share shares[n],
#                           uint8_t fails[n],
#                           uint8_t *fails_len);
def dkg_verify_commitments(n: int, t: int, self: int,
                           commitments : bytes_list_t,
                           shares: bytes_list_t) -> bytes:
    if n < t:
        raise ValueError("t cannot be bigger than n")
    if t < 2:
        raise ValueError("t must be bigger than 1")
    if self < 1 or self > n:
        raise ValueError("self must 1 <= self <= n")
    if len(commitments) != n*t*pysodium.crypto_core_ristretto255_BYTES:
        raise ValueError(f"signed_commitments must be {n*t*pysodium.crypto_core_ristretto255_BYTES} bytes is instead: {len(commitments)}")
    shares = b''.join(shares)
    if len(shares) != n*TOPRF_Share_BYTES:
        raise ValueError(f"shares must be {TOPRF_Share_BYTES*n} bytes is instead {len(shares)}")

    shares = ctypes.create_string_buffer(shares)
    fails = ctypes.create_string_buffer(n)
    fails_len = ctypes.c_uint8()
    __check(liboprf.dkg_verify_commitments(n, t, self,
                                           commitments, shares,
                                           fails, ctypes.byref(fails_len)))
    return fails[:fails_len.value]

#void dkg_finish(const uint8_t n,
#                const TOPRF_Share shares[n],
#                const uint8_t self,
#                TOPRF_Share *xi);
def dkg_finish(n: int, shares: List[bytes], self: int, ) -> bytes:
    if self < 1 or self > n:
        raise ValueError("self must 1 <= self <= n")
    shares = b''.join(shares)
    if len(shares) != n*TOPRF_Share_BYTES:
        raise ValueError(f"shares must be {TOPRF_Share_BYTES*n} bytes is instead {len(shares)}")

    shares = ctypes.create_string_buffer(shares)

    xi = ctypes.create_string_buffer(TOPRF_Share_BYTES)
    xi[0]=self

    liboprf.dkg_finish(n, shares, self, xi)
    return xi.raw

#void dkg_reconstruct(const size_t response_len,
#                     const TOPRF_Share responses[response_len][2],
#                     uint8_t result[crypto_scalarmult_ristretto255_BYTES]);
def dkg_reconstruct(responses) -> bytes_list_t:
    rlen = len(responses)
    responses = ctypes.create_string_buffer(b''.join(responses))
    result = ctypes.create_string_buffer(pysodium.crypto_core_ristretto255_BYTES)

    liboprf.dkg_reconstruct(rlen, responses, result)
    return result.raw

tpdkg_sessionid_SIZE=32
tpdkg_msg0_SIZE = 179 # ( sizeof(TP_DKG_Message)                       \
                      # + crypto_generichash_BYTES/*dst*/              \
                      # + 2 /*n,t*/                                    \
                      # + crypto_sign_PUBLICKEYBYTES /* tp_sign_pk */)
tpdkg_msg8_SIZE = 258 # (sizeof(TP_DKG_Message) /* header */                             \
                      #  + noise_xk_handshake3_SIZE /* 4th&final noise handshake */      \
                      #  + sizeof(TOPRF_Share) /* msg: the noise_xk wrapped share */     \
                      #  + crypto_secretbox_xchacha20poly1305_MACBYTES /* mac of msg */  \
                      #  + crypto_auth_hmacsha256_BYTES /* key-committing mac over msg*/ )
tpdkg_max_err_SIZE = 128

class TP_DKG_Cheater(ctypes.Structure):
    _fields_ = [('step',             ctypes.c_int),
                ('error',            ctypes.c_int),
                ('peer',             ctypes.c_uint8),
                ('other_peer',       ctypes.c_uint8),
                ('invalid_index',    ctypes.c_int),
                ]

#int tpdkg_start_tp(TP_DKG_TPState *ctx, const uint64_t ts_epsilon,
#             const uint8_t n, const uint8_t t,
#             const char *proto_name, const size_t proto_name_len,
#             const size_t msg0_len, TP_DKG_Message *msg0);
#
# also wraps conveniently:
#
# void tpdkg_tp_set_bufs(TP_DKG_TPState *ctx,
#                  uint8_t (*commitments)[][crypto_core_ristretto255_BYTES],
#                  uint16_t (*complaints)[],
#                  uint8_t (*suspicious)[],
#                  uint8_t (*tp_peers_sig_pks)[][crypto_sign_PUBLICKEYBYTES],
#                  uint8_t (*peer_lt_pks)[][crypto_sign_PUBLICKEYBYTES],
#                  uint64_t (*last_ts)[]);
def tpdkg_start_tp(n, t, ts_epsilon, proto_name, peer_lt_pks):
    b = ctypes.create_string_buffer(liboprf.tpdkg_tpstate_size()+32)
    b_addr = ctypes.addressof(b)
    s_addr = b_addr + (b_addr % 32)
    state = ctypes.c_void_p(s_addr)
    if state.value % 32 != 0:
      raise ValueError("cannot align at 32bytes the TP_DKG_PeerState struct")

    msg = ctypes.create_string_buffer(tpdkg_msg0_SIZE)
    __check(liboprf.tpdkg_start_tp(state, ctypes.c_uint64(ts_epsilon), ctypes.c_uint8(n), ctypes.c_uint8(t), proto_name, ctypes.c_size_t(len(proto_name)), ctypes.c_size_t(len(msg.raw)), msg))

    peers_sig_pks = ctypes.create_string_buffer(n*pysodium.crypto_sign_PUBLICKEYBYTES)
    commitments = ctypes.create_string_buffer(n*t*pysodium.crypto_core_ristretto255_BYTES)
    complaints = ctypes.create_string_buffer(n*n*2)
    noisy_shares = ctypes.create_string_buffer(n*n*tpdkg_msg8_SIZE)
    cheaters = (TP_DKG_Cheater * (t*t - 1))()
    peer_lt_pks = b''.join(peer_lt_pks)
    last_ts = (ctypes.c_uint64 * n)()

    liboprf.tpdkg_tp_set_bufs(state,
                              ctypes.byref(commitments),
                              ctypes.byref(complaints),
                              ctypes.byref(noisy_shares),
                              ctypes.byref(cheaters),
                              len(cheaters),
                              ctypes.byref(peers_sig_pks),
                              peer_lt_pks,
                              ctypes.byref(last_ts))

    # we need to keep these arrays around, otherwise the gc eats them up.
    ctx = (state, cheaters, peers_sig_pks, commitments, complaints, noisy_shares, peer_lt_pks, last_ts, b)

    return ctx, msg.raw


#size_t tpdkg_tp_input_size(const TP_DKG_TPState *ctx);
def tpdkg_tp_input_size(ctx):
   return liboprf.tpdkg_tp_input_size(ctx[0])

#int tpdkg_tp_input_sizes(const TP_DKG_TPState *ctx, size_t *sizes);
def tpdkg_tp_input_sizes(ctx):
   sizes = (ctypes.c_size_t * tpdkg_tpstate_n(ctx))()
   ret = liboprf.tpdkg_tp_input_sizes(ctx[0], ctypes.byref(sizes))
   return ret, [x for x in sizes]

#size_t tpdkg_tp_output_size(const TP_DKG_TPState *ctx);
def tpdkg_tp_output_size(ctx):
   return liboprf.tpdkg_tp_output_size(ctx[0])

#int tpdkg_tp_next(TP_DKG_TPState *ctx, const uint8_t *input, const size_t input_len, uint8_t *output, const size_t output_len);
def tpdkg_tp_next(ctx, msg):
    input_len = tpdkg_tp_input_size(ctx)
    if len(msg) != input_len: raise ValueError(f"input msg is invalid size: {len(msg)}B must be: {input_len}B")
    output_len = tpdkg_tp_output_size(ctx)
    output = ctypes.create_string_buffer(output_len)
    __check(liboprf.tpdkg_tp_next(ctx[0], msg, ctypes.c_size_t(input_len), output, ctypes.c_size_t(output_len)))
    return output

#int tpdkg_tp_peer_msg(const TP_DKG_TPState *ctx, const uint8_t *base, const size_t base_size, const uint8_t peer, const uint8_t **msg, size_t *len);
def tpdkg_tp_peer_msg(ctx, base, peer):
    msg = ctypes.POINTER(ctypes.c_char)()
    size = ctypes.c_size_t()
    __check(liboprf.tpdkg_tp_peer_msg(ctx[0], base, len(base.raw), peer, ctypes.byref(msg), ctypes.byref(size)))
    msg = b''.join([msg[i] for i in range(size.value)])
    return msg

#int tpdkg_tp_not_done(const TP_DKG_TPState *tp);
def tpdkg_tp_not_done(ctx):
    return liboprf.tpdkg_tp_not_done(ctx[0]) == 1

def tpdkg_get_cheaters(ctx):
    cheats = []
    cheaters = set()
    for i in range(tpdkg_tpstate_cheater_len(ctx)):
        err = ctypes.create_string_buffer(tpdkg_max_err_SIZE)
        p = liboprf.tpdkg_cheater_msg(ctypes.byref(ctx[1][i]), err, tpdkg_max_err_SIZE)
        if 0 >= p > tpdkg_tpstate_n(ctx):
            print(f"invalid cheater index: {p}, skipping this entry")
            continue
        cheaters.add(p)
        cheats.append((p, err.raw[:err.raw.find(b'\x00')].decode('utf8')))
    return cheaters, cheats

liboprf.tpdkg_peerstate_n.restype = ctypes.c_uint8
def tpdkg_peerstate_n(ctx):
    return liboprf.tpdkg_peerstate_n(ctx[0])
liboprf.tpdkg_peerstate_t.restype = ctypes.c_uint8
def tpdkg_peerstate_t(ctx):
    return liboprf.tpdkg_peerstate_t(ctx[0])
liboprf.tpdkg_peerstate_sessionid.restype = ctypes.POINTER(ctypes.c_uint8)
def tpdkg_peerstate_sessionid(ctx):
    ptr = liboprf.tpdkg_peerstate_sessionid(ctx[0])
    return bytes(ptr[i] for i in range(tpdkg_sessionid_SIZE))
liboprf.tpdkg_peerstate_lt_sk.restype = ctypes.POINTER(ctypes.c_uint8)
def tpdkg_peerstate_lt_sk(ctx):
    ptr = liboprf.tpdkg_peerstate_lt_sk(ctx[0])
    return bytes(ptr[i] for i in range(pysodium.crypto_sign_SECRETKEYBYTES))
liboprf.tpdkg_peerstate_share.restype = ctypes.POINTER(ctypes.c_uint8)
def tpdkg_peerstate_share(ctx):
    ptr = liboprf.tpdkg_peerstate_share(ctx[0])
    return bytes(ptr[i] for i in range(TOPRF_Share_BYTES))
def tpdkg_peerstate_step(ctx):
    return liboprf.tpdkg_peerstate_step(ctx[0])

liboprf.tpdkg_tpstate_n.restype = ctypes.c_uint8
def tpdkg_tpstate_n(ctx):
    return liboprf.tpdkg_tpstate_n(ctx[0])
liboprf.tpdkg_tpstate_t.restype = ctypes.c_uint8
def tpdkg_tpstate_t(ctx):
    return liboprf.tpdkg_tpstate_t(ctx[0])
liboprf.tpdkg_tpstate_cheater_len.restype = ctypes.c_size_t
def tpdkg_tpstate_cheater_len(ctx):
    return liboprf.tpdkg_tpstate_cheater_len(ctx[0])
liboprf.tpdkg_tpstate_sessionid.restype = ctypes.POINTER(ctypes.c_uint8)
def tpdkg_tpstate_sessionid(ctx):
    ptr = liboprf.tpdkg_tpstate_sessionid(ctx[0])
    return bytes(ptr[i] for i in range(tpdkg_sessionid_SIZE))
def tpdkg_tpstate_step(ctx):
    return liboprf.tpdkg_tpstate_step(ctx[0])

#int tpdkg_start_peer(TP_DKG_PeerState *ctx, const uint64_t ts_epsilon,
#               const uint8_t peer_lt_sk[crypto_sign_SECRETKEYBYTES],
#               const TP_DKG_Message *msg0);
#
# also wraps conveniently
#
#void tpdkg_peer_set_bufs(TP_DKG_PeerState *ctx,
#                         uint8_t (*peers_sig_pks)[][crypto_sign_PUBLICKEYBYTES],
#                         uint8_t (*peers_noise_pks)[][crypto_scalarmult_BYTES],
#                         Noise_XK_session_t *(*noise_outs)[],
#                         Noise_XK_session_t *(*noise_ins)[],
#                         TOPRF_Share (*shares)[],
#                         TOPRF_Share (*xshares)[],
#                         uint8_t (*commitments)[][crypto_core_ristretto255_BYTES],
#                         uint16_t (*complaints)[],
#                         uint8_t (*my_complaints)[]);
def tpdkg_peer_start(ts_epsilon, peer_lt_sk, msg0):
    b = ctypes.create_string_buffer(liboprf.tpdkg_peerstate_size()+32)
    b_addr = ctypes.addressof(b)
    s_addr = b_addr + (b_addr % 32)
    state = ctypes.c_void_p(s_addr)
    if state.value % 32 != 0:
      raise ValueError("cannot align at 32bytes the TP_DKG_PeerState struct")

    __check(liboprf.tpdkg_start_peer(state, ctypes.c_uint64(ts_epsilon), peer_lt_sk, msg0))

    n = tpdkg_peerstate_n([state])
    t = tpdkg_peerstate_t([state])

    peers_sig_pks = ctypes.create_string_buffer(b"peer_sig_pks", n * pysodium.crypto_sign_PUBLICKEYBYTES)
    peers_noise_pks = ctypes.create_string_buffer(b"peer_noise_pks", n * pysodium.crypto_scalarmult_BYTES)
    noise_outs = (ctypes.c_void_p * n)()
    noise_ins = (ctypes.c_void_p * n)()
    shares = ctypes.create_string_buffer(n * TOPRF_Share_BYTES)
    xshares = ctypes.create_string_buffer(n * TOPRF_Share_BYTES)
    commitments = ctypes.create_string_buffer(n * t * pysodium.crypto_core_ristretto255_BYTES)
    complaints = ctypes.create_string_buffer(n * n * 2)
    my_complaints = ctypes.create_string_buffer(n)
    last_ts = (ctypes.c_uint64 * n)()
    liboprf.tpdkg_peer_set_bufs(state,
                                ctypes.byref(peers_sig_pks),
                                ctypes.byref(peers_noise_pks),
                                noise_outs,
                                noise_ins,
                                ctypes.byref(shares),
                                ctypes.byref(xshares),
                                ctypes.byref(commitments),
                                ctypes.byref(complaints),
                                ctypes.byref(my_complaints),
                                ctypes.byref(last_ts))

    # we need to keep these arrays around, otherwise the gc eats them up.
    ctx = (state, peers_sig_pks, peers_noise_pks, noise_outs, noise_ins, shares, xshares, commitments, complaints, my_complaints, b, last_ts)
    return ctx

#size_t tpdkg_peer_input_size(const TP_DKG_PeerState *ctx);
def tpdkg_peer_input_size(ctx):
   return liboprf.tpdkg_peer_input_size(ctx[0])

#size_t tpdkg_peer_output_size(const TP_DKG_PeerState *ctx);
def tpdkg_peer_output_size(ctx):
   return liboprf.tpdkg_peer_output_size(ctx[0])

#int tpdkg_peer_next(TP_DKG_PeerState *ctx, const uint8_t *input, const size_t input_len, uint8_t *output, const size_t output_len);
def tpdkg_peer_next(ctx, msg):
    input_len = tpdkg_peer_input_size(ctx)
    if len(msg) != input_len: raise ValueError(f"input msg is invalid size: {len(msg)}B must be: {input_len}B")
    output_len = tpdkg_peer_output_size(ctx)
    output = ctypes.create_string_buffer(output_len)
    __check(liboprf.tpdkg_peer_next(ctx[0], msg, ctypes.c_size_t(input_len), output, ctypes.c_size_t(output_len)))
    return output.raw

#int tpdkg_peer_not_done(const TP_DKG_PeerState *peer);
def tpdkg_peer_not_done(ctx):
    return liboprf.tpdkg_peer_not_done(ctx[0]) == 1

#void tpdkg_peer_free(TP_DKG_PeerState *ctx);
def tpdkg_peer_free(ctx):
    liboprf.tpdkg_peer_free(ctx[0])



#!/usr/bin/env python3
import unittest
import pyoprf, pysodium, ctypes
from binascii import unhexlify
from itertools import combinations

class TestEndToEnd(unittest.TestCase):
  def test_cfrg_irtf(self):
    """CFRG/IRTF spec compliant run"""
    # Alice blinds the input "test"
    r, alpha = pyoprf.blind(b"test")
    # Bob generates a "secret" key
    k = pyoprf.keygen()
    # Bob evaluates Alices blinded value with it's key
    beta = pyoprf.evaluate(k, alpha)
    # Alice unblinds Bobs evaluation
    N = pyoprf.unblind(r, beta)
    # Alice finalizes the calculation
    y = pyoprf.finalize(b"test", N)
    # rerun and assert that oprf(k,"test") equals all runs
    r, alpha = pyoprf.blind(b"test")
    beta = pyoprf.evaluate(k, alpha)
    N = pyoprf.unblind(r, beta)
    y2 = pyoprf.finalize(b"test", N)
    self.assertEqual(y, y2)

  def test_cfrg_irtf_testvec1(self):
    """IRTF/CFRG testvector 1"""
    x = unhexlify("00")
    k = unhexlify("5ebcea5ee37023ccb9fc2d2019f9d7737be85591ae8652ffa9ef0f4d37063b0e")
    out=unhexlify("527759c3d9366f277d8c6020418d96bb393ba2afb20ff90df23fb7708264e2f3ab9135e3bd69955851de4b1f9fe8a0973396719b7912ba9ee8aa7d0b5e24bcf6")

    r, alpha = pyoprf.blind(x)
    beta = pyoprf.evaluate(k, alpha)
    N = pyoprf.unblind(r, beta)
    y = pyoprf.finalize(x, N)
    self.assertEqual(y,out)

  def test_cfrg_irtf_testvec2(self):
    """IRTF/CFRG testvector 2"""
    x=unhexlify("5a5a5a5a5a5a5a5a5a5a5a5a5a5a5a5a5a")
    k = unhexlify("5ebcea5ee37023ccb9fc2d2019f9d7737be85591ae8652ffa9ef0f4d37063b0e")
    out=unhexlify("f4a74c9c592497375e796aa837e907b1a045d34306a749db9f34221f7e750cb4f2a6413a6bf6fa5e19ba6348eb673934a722a7ede2e7621306d18951e7cf2c73")
    r, alpha = pyoprf.blind(x)
    beta = pyoprf.evaluate(k, alpha)
    N = pyoprf.unblind(r, beta)
    y = pyoprf.finalize(x, N)
    self.assertEqual(y, out)

  def test_hashDH_update(self):
    """HashDH with update example"""
    # Alice blinds the input "test"
    r, alpha = pyoprf.blind(b"test")
    # Bob generates a "secret" key
    k = pyoprf.keygen()

    # Bob evaluates Alices blinded value with it's key
    beta = pyoprf.evaluate(k, alpha)

    # Alice unblinds Bobs evaluation
    N = pyoprf.unblind(r, beta)

    # Bob updates his key, by generating delta
    delta = pysodium.crypto_core_ristretto255_scalar_random()

    k2 = pysodium.crypto_core_ristretto255_scalar_mul(k, delta)

    # Alice updates her previous calculation of N with delta
    N2 = pysodium.crypto_scalarmult_ristretto255(delta, N)

    # rerun hashDH to verify if N2 is equal with a full run
    r, alpha = pyoprf.blind(b"test")
    beta = pyoprf.evaluate(k2, alpha)
    N2_ = pyoprf.unblind(r, beta)
    self.assertEqual(N2, N2_)


  def test_toprf_sss(self):
    """tOPRF (hashDH), (3,5), with centrally shared key interpolation at client"""
    k2 = pyoprf.keygen()
    shares = pyoprf.create_shares(k2, 5, 3)
    r, alpha = pyoprf.blind(b"test")
    #print(' '.join(s.hex() for s in shares))
    # we reuse values from te previous test
    betas = tuple(s[:1]+pyoprf.evaluate(s[1:], alpha) for s in shares)
    #print(''.join(b.hex() for b in betas))

    beta = pyoprf.thresholdmult(betas)
    Nt = pyoprf.unblind(r, beta)

    beta = pyoprf.evaluate(k2, alpha)
    N2 = pyoprf.unblind(r, beta)
    self.assertEqual(N2, Nt)

  def test_toprf_tcombine(self):
    """tOPRF (hashDH), (3,5), with centrally shared key interpolation at servers"""
    k2 = pyoprf.keygen()
    shares = pyoprf.create_shares(k2, 5, 3)
    r, alpha = pyoprf.blind(b"test")

    indexes=(4,2,1)
    betas = tuple(pyoprf.threshold_evaluate(shares[i-1], alpha, i, indexes) for i in indexes)

    beta = pyoprf.threshold_combine(betas)

    beta = pyoprf.evaluate(k2, alpha)
    Nt = pyoprf.unblind(r, beta)

    Nt2 = pyoprf.unblind(r, beta)
    self.assertEqual(Nt, Nt2)

  def test_raw_dkg(self):
    """naked Distributed KeyGen (3,5)"""
    n = 5
    t = 3
    mailboxes=[[] for _ in range(n)]
    commitments=[]
    for _ in range(n):
        coms, shares = pyoprf.dkg_start(n,t)
        commitments.append(coms)
        for i,s in enumerate(shares):
            mailboxes[i].append(s)

    commitments=b''.join(commitments)

    shares = []
    for i in range(n):
       fails = pyoprf.dkg_verify_commitments(n,t,i+1,
                                             commitments,
                                             mailboxes[i])
       if len(fails) > 0:
           for fail in fails:
               print(f"fail: peer {fail}")
           raise ValueError("failed to verify contributions, aborting")
       xi = pyoprf.dkg_finish(n, mailboxes[i], i+1)
       #print(i, xi.hex(), x_i.hex())
       shares.append(xi)

    # test if the final shares all reproduce the same shared `secret`
    v0 = pyoprf.thresholdmult([bytes([i+1])+pysodium.crypto_scalarmult_ristretto255_base(shares[i][1:]) for i in (0,1,2)])
    for peers in combinations(range(1,5), 3):
      v1 = pyoprf.thresholdmult([bytes([i+1])+pysodium.crypto_scalarmult_ristretto255_base(shares[i][1:]) for i in peers])
      self.assertEqual(v0, v1)

    secret = pyoprf.dkg_reconstruct(shares[:t])
    #print("secret", secret.hex())
    self.assertEqual(v0, pysodium.crypto_scalarmult_ristretto255_base(secret))

  def test_explicit_3hashtdh(self):
    """toprf based on 2024/1455 [JSPPJ24] https://eprint.iacr.org/2024/1455
       using explicit implementation of 3hashtdh"""

    print("tOPRF (3hashTDH), (3,5), with centrally shared key interpolation at client")
    k2 = pyoprf.keygen()
    shares = pyoprf.create_shares(k2, 5, 3)
    zero_shares = pyoprf.create_shares(bytes([0]*32), 5, 3)

    r, alpha = pyoprf.blind(b"test")

    ssid_S = pysodium.randombytes(32)
    betas = []
    for k, z in zip(shares,zero_shares):
        h2 = pyoprf.evaluate(
            z[1:],
            pysodium.crypto_core_ristretto255_from_hash(pysodium.crypto_generichash(ssid_S + alpha, outlen=64)),
            )
        beta = pyoprf.evaluate(k[1:], alpha)
        betas.append(k[:1]+pysodium.crypto_core_ristretto255_add(beta, h2))

    # normal 2hashdh(k2,"test")
    beta = pyoprf.evaluate(k2, alpha)
    Nt0 = pyoprf.unblind(r, beta)
    for peers in combinations(betas, 3):
        beta = pyoprf.thresholdmult(betas[:3])
        Nt1 = pyoprf.unblind(r, beta)
        self.assertEqual(Nt0, Nt1)

  def test_native_3hashtdh(self):
    """toprf based on 2024/1455 [JSPPJ24] https://eprint.iacr.org/2024/1455
       using libopr native implementation of 3hashtdh
       tOPRF (3hashTDH), (3,5), with centrally shared key interpolation at client"""
    k2 = pyoprf.keygen()
    shares = pyoprf.create_shares(k2, 5, 3)
    zero_shares = pyoprf.create_shares(bytes([0]*32), 5, 3)

    r, alpha = pyoprf.blind(b"test")

    ssid_S = pysodium.randombytes(32)
    betas = []
    for k, z in zip(shares,zero_shares):
        betas.append(pyoprf._3hashtdh(k, z, alpha, ssid_S))

    beta = pyoprf.evaluate(k2, alpha)
    Nt0 = pyoprf.unblind(r, beta)
    for peers in combinations(betas, 3):
        beta = pyoprf.thresholdmult(betas[:3])
        Nt1 = pyoprf.unblind(r, beta)
        self.assertEqual(Nt0, Nt1)

  def test_tp_dkg(self):
    """Trusted Party Distributed KeyGeneration"""
    n = 5
    t = 3
    ts_epsilon = 5

    # enable verbose logging for tp-dkg
    #libc = ctypes.cdll.LoadLibrary('libc.so.6')
    #cstderr = ctypes.c_void_p.in_dll(libc, 'stderr')
    #log_file = ctypes.c_void_p.in_dll(pyoprf.liboprf,'log_file')
    #log_file.value = cstderr.value

    # create some long-term keypairs
    peer_lt_pks = []
    peer_lt_sks = []
    for _ in range(n):
        pk, sk = pysodium.crypto_sign_keypair()
        peer_lt_pks.append(pk)
        peer_lt_sks.append(sk)

    # initialize the TP and get the first message
    tp, msg0 = pyoprf.tpdkg_start_tp(n, t, ts_epsilon, "pyoprf tpdkg test", peer_lt_pks)

    print(f"\nn: {pyoprf.tpdkg_tpstate_n(tp)}, t: {pyoprf.tpdkg_tpstate_t(tp)}, sid: {bytes(c for c in pyoprf.tpdkg_tpstate_sessionid(tp)).hex()}")

    # initialize all peers with the 1st message from TP

    peers=[]
    for i in range(n):
        peer = pyoprf.tpdkg_peer_start(ts_epsilon, peer_lt_sks[i], msg0)
        peers.append(peer)

    for i in range(n):
        self.assertEqual(pyoprf.tpdkg_peerstate_sessionid(peers[i]), pyoprf.tpdkg_tpstate_sessionid(tp))
        self.assertEqual(peer_lt_sks[i], pyoprf.tpdkg_peerstate_lt_sk(peers[i]))

    peer_msgs = []
    while pyoprf.tpdkg_tp_not_done(tp):
        ret, sizes = pyoprf.tpdkg_tp_input_sizes(tp)
        # peer_msgs = (recv(size) for size in sizes)
        msgs = b''.join(peer_msgs)

        cur_step = pyoprf.tpdkg_tpstate_step(tp)
        try:
          tp_out = pyoprf.tpdkg_tp_next(tp, msgs)
          #print(f"tp: msg[{tp[0].step}]: {tp_out.raw.hex()}")
        except Exception as e:
          cheaters, cheats = pyoprf.tpdkg_get_cheaters(tp)
          print(f"Warning during the distributed key generation the peers misbehaved: {sorted(cheaters)}")
          for k, v in cheats:
              print(f"\tmisbehaving peer: {k} was caught: {v}")
          raise ValueError(f"{e} | tp step {cur_step}")

        peer_msgs = []
        while(len(b''.join(peer_msgs))==0 and pyoprf.tpdkg_peer_not_done(peers[0])):
            for i in range(n):
                if(len(tp_out)>0):
                    msg = pyoprf.tpdkg_tp_peer_msg(tp, tp_out, i)
                    #print(f"tp -> peer[{i+1}] {msg.hex()}")
                else:
                    msg = ''
                out = pyoprf.tpdkg_peer_next(peers[i], msg)
                if(len(out)>0):
                    peer_msgs.append(out)
                    #print(f"peer[{i+1}] -> tp {peer_msgs[-1].hex()}")
            tp_out = ''

    # we are done, let's check the shares

    shares = [pyoprf.tpdkg_peerstate_share(peers[i]) for i in range(n)]
    for i, share in enumerate(shares):
        print(f"share[{i+1}] {share.hex()}")

    v0 = pyoprf.thresholdmult([bytes([i+1])+pysodium.crypto_scalarmult_ristretto255_base(shares[i][1:]) for i in (0,1,2)])
    for peers_idxs in combinations(range(1,5), 3):
      v1 = pyoprf.thresholdmult([bytes([i+1])+pysodium.crypto_scalarmult_ristretto255_base(shares[i][1:]) for i in peers_idxs])
      self.assertEqual(v0, v1)

    secret = pyoprf.dkg_reconstruct(shares[:t])
    #print("secret", secret.hex())
    self.assertEqual(v0, pysodium.crypto_scalarmult_ristretto255_base(secret))

    # clean up allocated buffers
    for i in range(n):
        pyoprf.tpdkg_peer_free(peers[i])



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2023, Marsiske Stefan 
# SPDX-License-Identifier: LGPL-3.0-or-later

import os
from setuptools import setup, find_packages


# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(name = 'pyoprf',
       version = '0.7.1',
       description = 'python bindings for liboprf',
       license = "LGPLv3",
       author = 'Stefan Marsiske',
       author_email = 'toprf@ctrlc.hu',
       url = 'https://github.com/stef/liboprf/python',
       long_description=read('README.md'),
       long_description_content_type="text/markdown",
       packages=find_packages(),
       install_requires = ("pysodium", "SecureString"),
       classifiers = ["Development Status :: 4 - Beta",
                      "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
                      "Topic :: Security :: Cryptography",
                      "Topic :: Security",
                   ],
       #ext_modules = [liboprf],
)



#!/usr/bin/env python

import ssl, socket, select
from binascii import a2b_base64

class Peer:
    def __init__(self, name, addr, type = "SSL", ssl_cert=None, timeout=5):
        self.name = name
        self.type = type    # currently only TCP or SSL over TCP, but
                            # could be others like dedicated NOISE_XK,
                            # or hybrid mceliece+x25519 over USB or
                            # even UART
        self.address = addr # Currently only TCP host:port as a tuple
        self.ssl_cert = ssl_cert
        self.timeout = timeout
        self.state = "new"
        self.fd = None

    def connect(self):
        if self.state == "connected":
            raise ValueError(f"{self.name} is already connected")

        if self.type not in {"SSL", "TCP"}:
            raise ValueError(f"Unsupported peer type: {self.type}")

        if self.type == "SSL":
           ctx = ssl.create_default_context()
           ctx.minimum_version = ssl.TLSVersion.TLSv1_2
           if(self.ssl_cert):
               ctx.load_verify_locations(self.ssl_cert) # only for dev, production system should use proper certs!
               ctx.check_hostname=False                 # only for dev, production system should use proper certs!
               ctx.verify_mode=ssl.CERT_NONE            # only for dev, production system should use proper certs!
           else:
               ctx.load_default_certs()
               ctx.verify_mode = ssl.CERT_REQUIRED
               ctx.check_hostname = True

        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.settimeout(self.timeout)
        if self.type == "SSL":
            self.fd = ctx.wrap_socket(s, server_hostname=self.address[0])
        try: self.fd.connect(self.address)
        except: return
        self.state="connected"

    def connected(self):
        return self.state == "connected"

    def read(self,size):
        if not self.connected():
            return None
            #raise ValueError(f"{self.name} cannot read, is not connected")

        res = []
        read = 0
        while read<size and (len(res)==0 or len(res[-1])!=0):
          res.append(self.fd.recv(size-read))
          read+=len(res[-1])

        if len(res[-1])==0 and read<size:
            self.state = 'disconnected'
            #raise ValueError(f"short read for {self.name}, only {len(b''.join(res))} instead of expected {size} bytes")
        return b''.join(res)

    def send(self, msg):
        if not self.connected():
            return
            #raise ValueError(f"{self.name} cannot write, is not connected")
        self.fd.sendall(msg)

    def close(self):
        if self.state == "closed": return
        if not self.connected():
            return
            #raise ValueError(f"{self.name} cannot close, is not connected")
        if self.fd and self.fd.fileno() != -1:
            try: self.fd.shutdown(socket.SHUT_RDWR)
            except OSError: pass
            self.fd.close()
            self.state = "closed"
        else:
            # closed by other end.
            self.state = "closed"

class Multiplexer:
    def __init__(self, peers):
        self.peers = [Peer(name
                           ,(p['host'],p['port'])
                           ,type=p.get("type", "SSL")
                           ,ssl_cert = p.get('ssl_cert')
                           ,timeout = p.get('timeout'))
                      for name, p in peers.items()]

    def __getitem__(self, idx):
        return self.peers[idx]

    def __iter__(self):
        for p in self.peers:
            yield p

    def __len__(self):
        return len(self.peers)

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, exception_traceback):
        if exception_type is not None:
            print("exception caught", exception_type, exception_value, exception_traceback)
        self.close()

    def connect(self):
       for p in self.peers:
           p.connect()

    def send(self, idx, msg):
        self.peers[idx].send(msg)

    def broadcast(self, msg):
      for p in self.peers:
        p.send(msg)

    def gather(self, expectedmsglen, n=None, proc=None, debug=False):
       fails = []
       if n is None:
           n=len(self.peers)
       responses={}
       for idx, peer in enumerate(self.peers):
           if not peer.connected(): continue
           pending = peer.fd.pending()
           if pending == expectedmsglen:
               pkt = peer.read(expectedmsglen)
               if debug: print(f"{idx} got pending {len(pkt)}")
               if pkt == b'\x00\x04fail':
                   responses[idx]=None
                   continue
               responses[idx]=pkt if not proc else proc(pkt)
           #if peer.fd.fileno() < 0:
           #  print(f"{peer} has negative fileno: {peer.fd.fileno()}")
           #elif pending != 0:
           #    print(f"wtf peer {peer.name} has {peer.fd.pending()} bytes pending, which is not equ {expectedmsglen}")
       while len(responses)<n:
          fds={x.fd.fileno(): (i, x) for i,x in enumerate(self.peers) if i not in responses and x.connected() and x.fd.fileno() >= 0}
          if not fds: raise ValueError("not enough peers left to get enough results")
          #print("select")
          r, _,_ =select.select(fds.keys(),[],[],2)
          #print("select done")
          if not r: continue
          #print("got r")
          for fd in r:
             idx = fds[fd][0]
             if idx in responses:
                continue
             #print(f"gathering {idx}")
             pkt = fds[fd][1].read(expectedmsglen)
             if pkt == b'\x00\x04fail':
                 responses[idx]=None
                 continue
             if debug: print(f"{idx} got response of {len(pkt)}")
             tmp = pkt if not proc else proc(pkt)
             if tmp is None: continue
             responses[idx]=tmp
       if set((tuple(e) if isinstance(e,list) else e) for e in responses.values())=={None}:
           raise ValueError("oracles failed")
       if None in responses.values():
           if debug: print(f"some reponses failed")
           #return {k:v for k,v in responses.items() if v is not None}
       return [responses.get(i,None) for i in range(len(self.peers))]
       #return responses

    def close(self):
      for p in self.peers:
        p.close()



#!/usr/bin/env python

import sys
from construct import *

tpdkg_msg = Struct(
    "signature" / Array(64, Byte),
    "msgno" / Int8ub,
    "size" / Int32ub,
    "sender" / Int8ub,
    "to" / Int8ub,
    "ts" / Timestamp(Int64ub, 1., 1970),
    "data" / Array(this.size - 79, Byte),
)

messages = GreedyRange(tpdkg_msg)

with open(sys.argv[1], 'rb') as fd:
    raw = fd.read()

while len(raw) > 0:
    print(raw[:83].hex())
    try:
        msg = tpdkg_msg.parse(raw)
        print(f"{str(msg.ts)[:-6]} msgno: {msg.msgno}, len: {msg.size}, from: {msg.sender}, to: {msg.to:x}, data {bytes(msg.data).hex()}")
        raw = raw[msg.size:]
    except:
        print(raw[65:69].hex())
        raw = raw[83:]



#!/usr/bin/env python

"""
Test for TP DKG wrapper of pyoprf/liboprf

  SPDX-FileCopyrightText: 2024, Marsiske Stefan
  SPDX-License-Identifier: LGPL-3.0-or-later

  Copyright (c) 2024, Marsiske Stefan.
  All rights reserved.

  This file is part of liboprf.

  liboprf is free software: you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public License
  as published by the Free Software Foundation, either version 3 of
  the License, or (at your option) any later version.

  liboprf is distributed in the hope that it will be
  useful, but WITHOUT ANY WARRANTY; without even the implied
  warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  See the GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License
  along with liboprf. If not, see <http://www.gnu.org/licenses/>.

"""

import pyoprf, pysodium, ctypes

n = 5
t = 3
ts_epsilon = 5

# enable verbose logging for tp-dkg
libc = ctypes.cdll.LoadLibrary('libc.so.6')
cstderr = ctypes.c_void_p.in_dll(libc, 'stderr')
log_file = ctypes.c_void_p.in_dll(pyoprf.liboprf,'log_file')
log_file.value = cstderr.value

# create some long-term keypairs
peer_lt_pks = []
peer_lt_sks = []
for _ in range(n):
    pk, sk = pysodium.crypto_sign_keypair()
    peer_lt_pks.append(pk)
    peer_lt_sks.append(sk)

# initialize the TP and get the first message
tp, msg0 = pyoprf.tpdkg_start_tp(n, t, ts_epsilon, "pyoprf tpdkg test", peer_lt_pks)

print(f"n: {pyoprf.tpdkg_tpstate_n(tp)}, t: {pyoprf.tpdkg_tpstate_t(tp)}, sid: {bytes(c for c in pyoprf.tpdkg_tpstate_sessionid(tp)).hex()}")

# initialize all peers with the 1st message from TP

peers=[]
for i in range(n):
    peer = pyoprf.tpdkg_peer_start(ts_epsilon, peer_lt_sks[i], msg0)
    peers.append(peer)

for i in range(n):
    assert(pyoprf.tpdkg_peerstate_sessionid(peers[i]) == pyoprf.tpdkg_tpstate_sessionid(tp))
    assert(peer_lt_sks[i] == pyoprf.tpdkg_peerstate_lt_sk(peers[i]))

peer_msgs = []
while pyoprf.tpdkg_tp_not_done(tp):
    ret, sizes = pyoprf.tpdkg_tp_input_sizes(tp)
    # peer_msgs = (recv(size) for size in sizes)
    msgs = b''.join(peer_msgs)

    cur_step = pyoprf.tpdkg_tpstate_step(tp)
    try:
      tp_out = pyoprf.tpdkg_tp_next(tp, msgs)
      #print(f"tp: msg[{tp[0].step}]: {tp_out.raw.hex()}")
    except Exception as e:
      cheaters, cheats = pyoprf.tpdkg_get_cheaters(tp)
      print(f"Warning during the distributed key generation the peers misbehaved: {sorted(cheaters)}")
      for k, v in cheats:
          print(f"\tmisbehaving peer: {k} was caught: {v}")
      raise ValueError(f"{e} | tp step {cur_step}")

    peer_msgs = []
    while(len(b''.join(peer_msgs))==0 and pyoprf.tpdkg_peer_not_done(peers[0])):
        for i in range(n):
            if(len(tp_out)>0):
                msg = pyoprf.tpdkg_tp_peer_msg(tp, tp_out, i)
                #print(f"tp -> peer[{i+1}] {msg.hex()}")
            else:
                msg = ''
            out = pyoprf.tpdkg_peer_next(peers[i], msg)
            if(len(out)>0):
                peer_msgs.append(out)
                #print(f"peer[{i+1}] -> tp {peer_msgs[-1].hex()}")
        tp_out = ''

# we are done, let's check the shares

shares = [pyoprf.tpdkg_peerstate_share(peers[i]) for i in range(n)]
for i, share in enumerate(shares):
    print(f"share[{i+1}] {share.hex()}")

v0 = pyoprf.thresholdmult([bytes([i+1])+pysodium.crypto_scalarmult_ristretto255_base(shares[i][1:]) for i in (0,1,2)])
v1 = pyoprf.thresholdmult([bytes([i+1])+pysodium.crypto_scalarmult_ristretto255_base(shares[i][1:]) for i in (2,0,3)])
assert v0 == v1
v2 = pyoprf.thresholdmult([bytes([i+1])+pysodium.crypto_scalarmult_ristretto255_base(shares[i][1:]) for i in (2,1,4)])
assert v0 == v2

secret = pyoprf.dkg_reconstruct(shares[:t])
#print("secret", secret.hex())
assert v0 == pysodium.crypto_scalarmult_ristretto255_base(secret)

# clean up allocated buffers
for i in range(n):
    pyoprf.tpdkg_peer_free(peers[i])
